@@ -1,4 +1,17 @@
-"""pytest unittests"""
+# Copyright 2017-2020 Kamil Sindi
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 import sys
 import pytest
 
@@ -19,18 +32,18 @@ def test_empty():
 
 def test_with_args_kwargs():
     class FooInterface(Interface):
-        def foo(self, a, b=1, *args, **kwargs):
+        def foo(self, a, *args, b=1, **kwargs):
             pass
 
     with pytest.raises(NotImplementedError):
         @implements(FooInterface)
         class FooImplementationFail:
-            def foo(self, a, b=7, *args):
+            def foo(self, a, *args, b=7):
                 pass
 
     @implements(FooInterface)
     class FooImplementationPass:
-        def foo(self, a, b=1, *args, **kwargs):
+        def foo(self, a, *args, b=1, **kwargs):
             pass
 
 
@@ -70,6 +83,19 @@ def test_property():
             pass
 
 
+def test_property_inverse():
+    class FooInterface(Interface):
+        def foo(self):
+            pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail:
+            @property
+            def foo(self):
+                pass
+
+
 def test_setters():
     class FooInterface(Interface):
         @property
@@ -252,12 +278,12 @@ def test_missing_property():
 
     with pytest.raises(NotImplementedError):    # missing method
         @implements(FooInterface)
-        class FooImplementationFail:
+        class FooImplementationFail1:           # skipcq: PYL-W0612
             pass
 
     with pytest.raises(NotImplementedError):    # missing property decorator
         @implements(FooInterface)
-        class FooImplementationFail:
+        class FooImplementationFail2:           # skipcq: PYL-W0612
             def foo(self):
                 pass
 
@@ -288,12 +314,143 @@ def test_bad_constructor():
 def test_static():
     class FooInterface(Interface):
         @staticmethod
-        def foo():
+        def foo(a, b, c):
             pass
 
     with pytest.raises(NotImplementedError):
         @implements(FooInterface)
-        class FooImplementation:
+        class FooImplementationFail1:           # skipcq: PYL-W0612
+            pass                    # missing foo
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail2:           # skipcq: PYL-W0612
+            # skipcq: PYL-E0213
+            def foo(a, b, c):       # missing staticmethod decorator
+                pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail3:           # skipcq: PYL-W0612
+            @classmethod            # classmethod instead of staticmethod
+            def foo(cls, a, b, c):  # decorator-check fails before signature
+                pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail4:           # skipcq: PYL-W0612
+            @staticmethod
+            def foo(m, n, o):       # staticmethod, but wrong signature
+                pass
+
+    @implements(FooInterface)
+    class FooImplementationPass:
+        @staticmethod
+        def foo(a, b, c):
+            pass
+
+
+def test_classmethods():
+    class FooInterface(Interface):
+        @classmethod
+        def foo(cls, a, b, c):
+            pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail1:           # skipcq: PYL-W0612
+            pass                    # missing foo
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail2:           # skipcq: PYL-W0612
+            # skipcq: PYL-E0213
+            def foo(cls, a, b, c):  # missing classmethod decorator
+                pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail3:           # skipcq: PYL-W0612
+            @staticmethod           # staticmethod instead of classmethod
+            def foo(a, b, c):       # decorator-check fails before signature
+                pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail4:           # skipcq: PYL-W0612
+            @classmethod
+            def foo(cls, m, n, o):   # classmethod, but wrong signature
+                pass
+
+    @implements(FooInterface)
+    class FooImplementationPass:
+        @classmethod
+        def foo(cls, a, b, c):
+            pass
+
+
+def test_classmethod_signature_match():
+    # For a classmethod, inspect.signature returns a signature with the first
+    # element (cls) stripped. A classmethod with signature (cls, a, b, c) has
+    # signature equivalence with a regular method with signature (a, b, c)
+    #
+    # Example:
+    from inspect import signature
+
+    class TestA:
+        @classmethod
+        def foo(cls, a, b, c):
+            pass
+
+    class TestB:
+        # skipcq: PYL-E0213
+        def foo(a, b, c):
+            pass
+
+    assert signature(TestA.foo) == signature(TestB.foo)
+
+    # The test below ensures that the above case is flagged
+    class FooInterface(Interface):
+        @classmethod
+        def foo(cls, a, b, c):
+            pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(FooInterface)
+        class FooImplementationFail:
+            # skipcq: PYL-E0213
+            def foo(a, b, c):
+                pass
+
+
+def test_staticmethod_classmethod_with_decorator():
+    class FooBarInterface(Interface):
+        @staticmethod
+        def foo(a, b, c):
+            pass
+
+        @classmethod
+        def bar(cls, a, b, c):
+            pass
+
+    import functools
+
+    def decorator(func):
+        @functools.wraps(func)
+        def inner(*args, **kwargs):
+            return func(*args, **kwargs)
+        return inner
+
+    @implements(FooBarInterface)
+    class FooBarImplementationPass:
+        @staticmethod
+        @decorator
+        def foo(a, b, c):
+            pass
+
+        @classmethod
+        @decorator
+        def bar(cls, a, b, c):
             pass
 
 
@@ -395,9 +552,10 @@ def test_interface_inheritance():
                 pass
 
     @implements(FooInterface)
-    class FooImplementationPass(object):
+    class FooImplementationPass:
         def foo(self):
             pass
+
         def bar(self):
             pass
 
@@ -433,17 +591,17 @@ def test_class_multiple_inheritance():
 
     # --------- IMPLEMENTATION -------------------------------------------
     #
-    class BaseFooImplementation(object):        # must get overridden
+    class BaseFooImplementation:        # must get overridden
         def foo(self, override, my, args):
             pass
 
     @implements(FooInterface)
     class FooImplementation(BaseFooImplementation):
-        def foo(self, final):
+        def foo(self, final):           # skipcq: PYL-W0221
             pass
 
     @implements(BarInterface)
-    class BarImplementation(object):
+    class BarImplementation:
         def bar(self, final):
             pass
 
@@ -493,27 +651,15 @@ def test_arg_type_annotation():
             pass
 
 
-def test_classmethods():
-    class FooInterface(Interface):
-        @classmethod
-        def foo(cls):
-            pass
-
-    with pytest.raises(NotImplementedError):
-        @implements(FooInterface)
-        class FooImplementation:
-            pass
-
-
 def test_other_decorator_compat():
     def decorator(cls):
-        class Wrapper(object):
+        class Wrapper:
             def __init__(self, *args):
                 self.wrapped = cls(*args)
 
             def __getattr__(self, name):
                 print('Getting the {} of {}'.format(name, self.wrapped))
-                return getattr(self.wrapped, name)
+                return getattr(self.wrapped, name, None)
 
         return Wrapper
 
@@ -524,7 +670,7 @@ def test_other_decorator_compat():
     with pytest.raises(NotImplementedError):
         @implements(FooInterface)
         @decorator
-        class FooImplementationFail(object):
+        class FooImplementationFail:
             def __init__(self, x, y):
                 self.x = x
                 self.y = y
@@ -534,7 +680,7 @@ def test_other_decorator_compat():
 
     @decorator
     @implements(FooInterface)
-    class FooImplementationPass(object):
+    class FooImplementationPass:
         def __init__(self, x, y):
             self.x = x
             self.y = y
@@ -550,11 +696,11 @@ def test_magic_methods():
 
     with pytest.raises(NotImplementedError):
         @implements(FooInterface)
-        class FooImplementationFail(object):
+        class FooImplementationFail:
             pass
 
     @implements(FooInterface)
-    class FooImplementationPass(object):
+    class FooImplementationPass:
         def __add__(self, other):
             pass
 
@@ -565,11 +711,11 @@ def test_attributes():
 
     with pytest.raises(NotImplementedError):
         @implements(FooInterface)
-        class FooImplementationFail(object):
+        class FooImplementationFail:
             pass
 
     @implements(FooInterface)
-    class FooImplementationPass(object):
+    class FooImplementationPass:
         a = 1
         b = 2
 
@@ -588,6 +734,70 @@ def test_async():
             pass
 
 
+def test_async_method():
+    class AsyncFooInterface:
+        async def foo(self):
+            pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(AsyncFooInterface)
+        class FooImplementationFail:                # skipcq: PYL-W0612
+            def foo(self):
+                pass
+
+    @implements(AsyncFooInterface)
+    class AsyncFooImplementation:                   # skipcq: PYL-W0612
+        async def foo(self):
+            pass
+
+
+def test_generator():
+    class GenFooInterface:
+        def foo(self):                              # skipcq: PYL-R0201
+            yield 1
+
+    with pytest.raises(NotImplementedError):
+        @implements(GenFooInterface)
+        class FooImplementationFail:                # skipcq: PYL-W0612
+            def foo(self):
+                pass
+
+    # must fail a generator which happens to be async
+    with pytest.raises(NotImplementedError):
+        @implements(GenFooInterface)
+        class AsyncGenFooImplementationFail:        # skipcq: PYL-W0612
+            async def foo(self):
+                yield 1
+
+    @implements(GenFooInterface)
+    class GenFooImplementation:                     # skipcq: PYL-W0612
+        def foo(self):                              # skipcq: PYL-R0201
+            yield 1
+
+
+def test_asyncgen_method():
+    class AsyncGenFooInterface:
+        async def foo(self):
+            yield 1
+
+    with pytest.raises(NotImplementedError):
+        @implements(AsyncGenFooInterface)
+        class AsyncFooImplementationFail:           # skipcq: PYL-W0612
+            async def foo(self):
+                pass
+
+    with pytest.raises(NotImplementedError):
+        @implements(AsyncGenFooInterface)
+        class GenFooImplementationFail:             # skipcq: PYL-W0612
+            def foo(self):                          # skipcq: PYL-R0201
+                yield 1
+
+    @implements(AsyncGenFooInterface)
+    class AsyncGenFooImplementation:                # skipcq: PYL-W0612
+        async def foo(self):
+            yield 1
+
+
 @py36
 def test_new_style_descriptors():
     class IntField:
@@ -600,7 +810,7 @@ def test_new_style_descriptors():
             instance.__dict__[self.name] = value
 
         def __set_name__(self, owner, name):
-            self.name = name
+            self.name = name                            # skipcq: PYL-W0201
 
     class FooInterface(Interface):
         int_field = IntField()
